% Statistical and Numerical Methods for Chemical Engineers 2023
% Exercise 3, assignment 2
clear all; close all; clc

%% Define constants
a = 49.46;
k0 = 2.17e20;
kc = 0.83;
eta = 0.33;
thetaC = 0.9;

%% Define functions for reaction rate and heat flow
k = @(t)k0.*exp(-a./t);
f = @(t)eta.*k(t)./(1+k(t)) + 1 - t + kc.*(thetaC-t);
df = @(t)eta*a/k0*exp(a./t)./(t.^2.*(1/k0*exp(a./t)+1).^2)-1-kc; % Derivative

%% Q2.1 Plot the heat flow against the temperature
t = linspace(0.9,1.25);
plot(t,f(t),'linewidth',2)
xlabel('Temperature')
ylabel('Heat flow')
yline(0) % display x axis on the plot

%% Q2.2 Determine the zeros of the heat flow function by chosing apropriate
% initial guesses
tic
[x1, xk1] = secantRoot(f,0.9);
toc
tic
[x2, xk2] = secantRoot(f,1.08);
toc
tic
[x3, xk3] = secantRoot(f,1.5);
toc
fprintf('Steady State at: %g; %g; %g with %g; %g; %g steps\n',x1,x2,x3,length(xk1),length(xk2),length(xk3))

% Q2.5 Determine the local convergence order and rate
deltaXs = log(abs(diff(xk1)));
X = deltaXs(1:end-1);
Y = deltaXs(2:end);
secantFit = polyfit(X,Y,1); % use built-in polynomial fit function to estimate y-intercept(C) and slope(q)
fprintf('Secant Method; Convergence order: %g; Convergence rate: %g\n',secantFit(1),exp(secantFit(2)));

%% Find out of convergence towards the three solutions
X = 0.01:0.01:2; %Vector of initial guess
X0 = zeros(size(X));
for i = 1:length(X)
    [X0(i),xvec]= secantRoot(f,X(i));
end
chg = find(diff(X0)>0.0001); % find positions of transition
figure
plot(X,X0,'-r');
xlabel('Initial guess')
ylabel('Zero')
hold on
line([X(chg(1)),X(chg(1))],[min(X0),max(X0)]);
line([X(chg(2)),X(chg(2))],[min(X0),max(X0)]);
axis('tight')

% Interpretation:
% 1) a very small epsilon would make the difference between x(k) and x(k-1)
% very small and the machine precision would make f'(x_k) a division with 0
% and not return a result. 

% 2)The convergence is driven by the slope. The two transition points
% are located at the two extreme points (slope = 0) and converge to
% the zeros in the direction of the slope from there




%% Newton method:
% Q2.3 Determine the zeros of the heat flow function by chosing apropriate
% initial guesses
tic
[x1, xk1] = newtonroot(f,df,0.9);
toc
tic
[x2, xk2] = newtonroot(f,df,1.08);
toc
tic
[x3, xk3] = newtonroot(f,df,1.5);
toc
fprintf('Steady State at: %g; %g; %g with %g; %g; %g steps\n',x1,x2,x3,length(xk1),length(xk2),length(xk3))


% Q2.5 Determine the local convergence order and rate
deltaXs = log(abs(diff(xk1)));
X = deltaXs(1:end-1);
Y = deltaXs(2:end);
newtonFit = polyfit(X,Y,1); % use built-in polynomial fit function to estimate y-intercept(C) and slope(q)
fprintf('Newton Method; Convergence order: %g; Convergence rate: %g\n',newtonFit(1),exp(newtonFit(2)));

%% Newton method with finite differences:
% Q2.4 Determine the zeros of the heat flow function by chosing apropriate
% initial guesses
h = 1e-6;
tic
[x1, xk1] = newtondiffroot(f,0.9,h);
toc
tic
[x2, xk2] = newtondiffroot(f,1.08,h);
toc
tic
[x3, xk3] = newtondiffroot(f,1.5,h);
toc
fprintf('Steady State at: %g; %g; %g with %g; %g; %g steps\n',x1,x2,x3,length(xk1),length(xk2),length(xk3))

h = 1e-16;
tic
[x1, xk1] = newtondiffroot(f,0.9,h);
toc
tic
[x2, xk2] = newtondiffroot(f,1.08,h);
toc
tic
[x3, xk3] = newtondiffroot(f,1.5,h);
toc
fprintf('Steady State at: %g; %g; %g with %g; %g; %g steps\n',x1,x2,x3,length(xk1),length(xk2),length(xk3))

% Q2.5 Determine the local convergence order and rate
deltaXs = log(abs(diff(xk1)));
X = deltaXs(1:end-1);
Y = deltaXs(2:end);
newtondiffFit = polyfit(X,Y,1); % use built-in polynomial fit function to estimate y-intercept(C) and slope(q)
fprintf('Finite Difference Approximation Method; Convergence order: %g; Convergence rate: %g\n',newtondiffFit(1),exp(newtondiffFit(2)));

%Q2.5 Conclusions:
%1) Convergence depends on the starting point for all methods.
%2) Convergence for Newton methods is faster (and less steps) than secant.
%3) Finite difference outperforms normal Newton method at h = 1e-6. Slow or
%not converging at 1e-16.

%% Function Implementation

% Secant Method
function [x,xvec] = secantRoot(f,x0)
%Q2.2 secantRoot approximates finds a zero of a function
%using the Secant Method

% Inputs:
% function handle f
% Initial guess x0 (scalar)

% Outputs:
% zero x (scalar)
% convergence history xvec (vector)

% Set initial values
x_0 = x0;
x_1 = (1+1e-5)*x0;
xvec(1) = x_0;
xvec(2) = x_1;
% Iterate until convergence is fulfilled
n = 0;
while abs(x_1-x_0) > 1e-8 && abs(f(x_1)) > 1e-6 && n < 1e5
    fprime = (f(x_1)-f(x_0))/(x_1-x_0);
    x_0 = x_1;
    x_1 = x_1 - f(x_1)./fprime;
    xvec = [xvec, x_1];
    n = n + 1;
end
% Root is final value
x = x_1;
end

% Newton Method
function [x,xvec] = newtonroot(f,df,x0)
%Q2.3 newtonroot finds the zero of a function
%using the Newton Method

% Inputs:
% function handle of function f
% function handle of derivative df
% Initial guess x0 (scalar)

% Outputs:
% zero x (scalar)
% convergence history xvec (vector)


n = 0;
x = x0;
dx = 1;
xvec = [x0];
% Loop until one of the criteria is no longer fulfilled
while abs(dx) > 1e-8 && abs(f(x)) > 1e-6 && n < 1e5
    % Calculate relevant information for method
    n = n+1;
    fx = f(x);
    Jx = df(x);
    dx = -fx/Jx;
    % Go on with the iteration
    x = x + dx;
    xvec = [xvec, x];
end
end

% Function approximation method
function [x,xvec] = newtondiffroot(f,x0,h)
%Q2.4 newtonroot finds the zero of a function
%using the Newton Method with forward difference approximation of
%derivative

% Inputs:
% function handle of function f
% Initial guess x0 (scalar)
% Stepsize for forward difference

% Outputs:
% zero x (scalar)
% convergence history xvec (vector)


n = 0;
x = x0;
dx = 1;
xvec = [x0];
% Loop until one of the criteria is no longer fulfilled
while abs(dx) > 1e-8 && abs(f(x)) > 1e-6 && n < 1e5
    % Calculate relevant information for method
    n = n+1;
    fx = f(x);
    Jx = (f(x+h)-f(x))/h;
    dx = -fx/Jx;
    % Go on with the iteration
    x = x + dx;
    xvec = [xvec, x];
end
end
